#include "i8080.h"
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <termios.h>

#define DISKINFO 0xFB65
void bios() {
  int straddr, i, fd, diskinfo, track, sector, dmaad, diskno;
  unsigned char c, buffer[128];
  struct termios settings, save_settings;

  switch (regPC) {
    case 0xFFE0:
      tcgetattr(0,&save_settings);
      settings = save_settings;
      settings.c_lflag &= ~(ECHO|ICANON|ISIG);
      settings.c_cc[VTIME] = 0;
      settings.c_cc[VMIN] = 1;
      settings.c_cc[VSTOP] = 0;
      settings.c_cc[VSTART] = 0;
      settings.c_cc[VSUSP] = 0;
      tcsetattr(0, TCSANOW, &settings);
      c = getchar();
      tcsetattr(0, TCSANOW, &save_settings);
      if (c == 0x0A) {
	c = 0x0D;
      }
      if (c == 0x10) {
        printf("\x1b[0;39m\n");
	exit(0);
      }
      regA = (c & 0x7F);
      break;
    case 0xFFE1:
      c = regA;
      putchar(c);
      break;
    case 0xFFE2:
      diskinfo = DISKINFO;
      track  = ram[diskinfo++] + (ram[diskinfo++] << 8);
      sector = ram[diskinfo++] + (ram[diskinfo++] << 8);
      dmaad  = ram[diskinfo++] + (ram[diskinfo++] << 8);
      diskno = ram[diskinfo];

      if (diskno == 0) fd = open("disk1.dat", O_RDONLY);
      else if (diskno == 1) fd = open("disk2.dat", O_RDONLY);
      else if (diskno == 2) fd = open("disk3.dat", O_RDONLY);
      else if (diskno == 3) fd = open("disk4.dat", O_RDONLY);

      if (fd == -1) {
        regA = 1;
        break;
      }
      if (lseek(fd, (track*64 + sector) * 128, SEEK_SET) == -1) {
        regA = 1;
        close(fd);
        break;
      }
      if (read(fd, buffer, 128) == -1) {
        regA = 1;
        close(fd);
        break;
      }
      for (i=0; i<128; i++) {
        ram[dmaad + i] = buffer[i];
      }
      regA = 0;
      close(fd);
      break;
    case 0xFFE3:
      diskinfo = DISKINFO;
      track  = ram[diskinfo++] + (ram[diskinfo++] << 8);
      sector = ram[diskinfo++] + (ram[diskinfo++] << 8);
      dmaad  = ram[diskinfo++] + (ram[diskinfo++] << 8);
      diskno = ram[diskinfo];

      if (diskno == 0) fd = open("disk1.dat", O_WRONLY | O_SYNC);
      else if (diskno == 1) fd = open("disk2.dat", O_WRONLY | O_SYNC);
      else if (diskno == 2) fd = open("disk3.dat", O_WRONLY | O_SYNC);
      else if (diskno == 3) fd = open("disk4.dat", O_WRONLY | O_SYNC);

      if (fd == -1) {
        regA = 1;
        break;
      }
      if (lseek(fd, (track*64 + sector) * 128, SEEK_SET) == -1) {
        regA = 1;
        close(fd);
        break;
      }
      for (i=0; i<128; i++) {
        buffer[i] = ram[dmaad + i];
      }
      if (write(fd, buffer, 128) == -1) {
        regA = 1;
        close(fd);
        break;
      }
      regA = 0;
      close(fd);
      break;
    default:
      break;
  }
}

int main(int argc, char *argv[]) {
  unsigned char opcode;

BOOT:
  regPC = 0xFA03;  // warm boot
  regSP = 0xFFFF;
  regA = regB = regC = regD = regE = regH = regL = 0;
  regF = 0b00000010;
  intr_flag = 0;

  FILE *fd = fopen("CPM.img", "r");
  if (!fd) {
    perror("can't open CPM.img");
    exit(EXIT_FAILURE);
  }
  int c;
  unsigned short i = 0xE400;
  while ((c = fgetc(fd)) != EOF) {
    ram[i++] = (unsigned char)(c & 0xFF);
  }
  fclose(fd);

  for(int i=0; i<4; i++) ram[0xFFE0 + i] = 0xC9;

  printf("\x1b[1;32m");
  printf("\e]12;green\a");
  printf("\nCP/M VERSION 2.2 EXTENSION 64k SYSTEM\n");
  printf("     Copyright (C) 1979 by Digital Research\n");

  for(;;) {
    opcode = fetch();
    exec(opcode);

    // CP/M cold boot
    if (regPC == 0xFA00) {
      goto BOOT;
    }
    // CP/M BIOS call
    if (regPC >= 0xFFE0) {
      bios();
    }
  }
}
